import argparse
import os
import logging
import sys

from snipty.base import Snipty
from . import __VERSION__

parser = argparse.ArgumentParser(prog='snipty', description='Minimalistic package manager for snippets.')
parser.add_argument('-q', '--quiet', action='count', default=0,
                    help='Give less output. Option is additive, and can be used up to 3 times (corresponding to '
                         'WARNING, ERROR, and CRITICAL logging levels).')
parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + __VERSION__,
                    help='Show version and exit.')

parser.add_argument('-p', '--path', default=os.environ.get('SNIPTY_ROOT_PATH', os.getcwd()),
                    metavar='<project root path>',
                    help='Project root path; default: SNIPTY_ROOT_PATH environment variable or current directory')

subparsers = parser.add_subparsers(title='Commands', dest='command')

parser_freeze = subparsers.add_parser('freeze', help='Freeze installed snippets')

parser_install = subparsers.add_parser('install', help='Install snippets')

parser_install.add_argument('-f', '--force', action='store_true',
                            help="Force installation even if snippet was already installed or path exists")

parser_install_group = parser_install.add_mutually_exclusive_group(required=True)

parser_install_group.add_argument('-r', '--requirement', type=argparse.FileType('rb', 0), metavar='<requirements file>',
                                  help='Install from the given requirements file.')

parser_install_group.add_argument('snippet_url', nargs='?', help='snippets url', metavar='<snippets url>')

parser_install.add_argument('snippet_name', nargs='?', metavar='<snippets name>',
                            help='snippet nape; can be a path')

logging.basicConfig(level=logging.INFO, format='%(message)s')
logger = logging.getLogger('snipty')
logger.setLevel(logging.INFO)


def install(args):
    """Calls snipty logic depending on arguments"""
    im = Snipty(project_root=args.path, force=args.force)
    if args.requirement:
        im.install_from_file(args.requirement)
    elif args.snippet_name and args.snippet_url:
        im.install_package(name=args.snippet_name, url=args.snippet_url)


def freeze(args):
    """Calls snipty logic for freeze"""
    Snipty(project_root=args.path).freeze()


if __name__ == '__main__':
    args = parser.parse_args()

    if args.command is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    if args.quiet == 1:
        logger.setLevel(logging.WARNING)

    elif args.quiet == 2:
        logger.setLevel(logging.ERROR)

    elif args.quiet >= 3:
        logger.setLevel(logging.CRITICAL)

    # Dispatch command
    {
        'install': install,
        'freeze': freeze
    }[args.command](args)



import json
import os
import re
import sys
from functools import wraps

import logging

from snipty.downloaders import BasicDownloader, BaseDownloader, DownloaderError, GhostbinDownloader

logger = logging.getLogger('snipty')


class NotExists(Exception):
    pass


def ensure_config_saved(f):
    @wraps(f)
    def wrapped(self, *args, **kwargs):
        try:
            f(self, *args, **kwargs)
        finally:
            self.store_config()

    return wrapped


class Snipty:
    """Manages whole process of tracking what is installed and calling specialized downloaders"""

    SUPPORTED_DOWNLOADERS = [GhostbinDownloader, BasicDownloader]

    def __init__(self, project_root, force=False):
        self.project_root = project_root
        self._config = None
        self.force = force

    def _dispatch_url(self, url) -> BaseDownloader:
        for downloader in self.SUPPORTED_DOWNLOADERS:
            if downloader.match(url):
                return downloader
        logger.error("Error: cannot find downloader for provided url {}".format(url))
        sys.exit(4)

    def _prepare_directory(self, root_path, package_dir):
        """Create a tree of directories and place __init__.py files"""
        full_path = os.path.join(root_path, package_dir)

        os.makedirs(full_path, exist_ok=True)

        while full_path != root_path:
            init_path = os.path.join(full_path, '__init__.py')
            if not os.path.exists(init_path):
                open(init_path, 'a').close()
            full_path = os.path.dirname(full_path)

    def _install_package(self, url, name):
        if not self.force and name in self.config(create=True)['packages_names']:
            logger.warning("Snippet '{}' has been already installed.".format(name))
            return

        if not self.force and url in self.config(create=True)['packages_urls']:
            logger.error(
                "Error: Snippet '{}' has been already from the same source {}.".format(
                    self.config(create=True)['packages_urls'][url], url))
            sys.exit(3)

        if not self.force and os.path.exists(os.path.join(self.project_root, name + '.py')) or os.path.exists(
                os.path.join(self.project_root, name)):
            logger.error("Error: Cannot install snippet '{}' because destination location already exists.".format(name))

        downloader_class = self._dispatch_url(url)

        try:
            tmp_path = downloader_class.download(url=url)
        except DownloaderError as e:
            logger.error("Error: Snippet {} cannot be installed - {}.".format(name, str(e)))
            sys.exit(6)

        # tmp_path can be a single file or directory (support for snippets containing many files)

        if os.path.isdir(tmp_path):
            package_dir = name
            package_name = None
        else:
            package_dir = os.path.dirname(name)
            package_name = os.path.basename(name)

        self._prepare_directory(self.project_root, package_dir)

        if package_name is not None:
            os.rename(tmp_path, os.path.join(self.project_root, package_dir, package_name + '.py'))
        else:
            # FIXME: move many files
            pass

        self.config(create=True)['packages_names'][name] = url
        self.config(create=True)['packages_urls'][url] = name

        logger.info('✔️ Snippet {} installed from {}'.format(name, url))

    @ensure_config_saved
    def install_package(self, url, name):
        self._install_package(url, name)

    @ensure_config_saved
    def install_from_file(self, requirements):
        for line_no, line in enumerate(requirements, 1):
            line = line.decode('utf-8').strip()

            # skip comments and empty lines
            if line.startswith('#') or not line:
                continue

            match = re.match('^(?P<name>.+)\s+from\s+(?P<url>https?://.*)$', line)
            if not match:
                logger.error('Error: Requirements file syntax error in line {}.'.format(line_no))

            self._install_package(url=match.group('url'), name=match.group('name'))

    def freeze(self):
        try:
            packages = self.config().get('packages_names', {})
            for package in sorted(packages):
                print(package, 'from', packages[package])

        except NotExists:
            logger.error('Error: Snipty was not used before in this project root path: {}'.format(self.project_root))
            sys.exit(1)

    @property
    def config_file_path(self):
        return os.path.join(self.project_root, '.snipty')

    def config(self, create=False):
        if self._config is None:
            if not os.path.exists(self.config_file_path):
                if not create:
                    raise NotExists

                self._store_config({})

            with open(self.config_file_path, 'r') as f:
                self._config = json.load(f)

            for key in ('packages_names', 'packages_urls'):
                if key not in self._config:
                    self._config[key] = {}

        return self._config

    def _store_config(self, data):
        with open(self.config_file_path, 'w') as f:
            json.dump(data, f, indent=2)

    def store_config(self):
        try:
            config = self.config()
            self._store_config(config)
        except NotExists:
            pass



from setuptools import setup
from snipty import __VERSION__

setup(
    name='snipty',
    version=__VERSION__,
    packages=['snipty'],
    python_requires='>3.6',
    url='',
    license='MIT',
    author='Kris Dorosz',
    author_email='cypreess@gmail.com',
    description='Minimalistic package manager for snippets.',
    install_requires=['requests>=2.18'],
    scripts=['bin/snipty']
)



import logging
import os
import tempfile

import requests

logger = logging.getLogger('snipty')


class DownloaderError(Exception):
    pass


class BaseDownloader:

    # @abstractmethod
    @classmethod
    def match(cls, url: str):
        """Check if given downloader can handle this request"""
        raise NotImplementedError

    # @abstractmethod
    @classmethod
    def download(cls, url: str) -> str:
        """Should get the raw snippet content from url and return a path to temporary file or directory"""
        raise NotImplementedError


class BasicDownloader(BaseDownloader):
    """
    Basic downloader will get any file that is plain/text type and HTTP status code was 200
    """

    ACCEPTED_CONTENT_TYPE = 'text/plain'
    ACCEPTED_HTTP_STATUS = 200

    @classmethod
    def match(cls, url: str):
        return True

    @classmethod
    def _fetch_file(self, url):
        with tempfile.NamedTemporaryFile(delete=False, dir=os.environ.get('SNIPTY_TMP')) as destination_file:
            response = requests.get(url)
            if response.status_code != BasicDownloader.ACCEPTED_HTTP_STATUS:
                raise DownloaderError('could not fetch {} (HTTP{})'.format(url, response.status_code))
            if not response.headers['content-type'].startswith(BasicDownloader.ACCEPTED_CONTENT_TYPE):
                raise DownloaderError('not a {} format.'.format(BasicDownloader.ACCEPTED_CONTENT_TYPE))

            for block in response.iter_content(1024):
                destination_file.write(block)

            return destination_file.name

    @classmethod
    def download(cls, url: str):
        return cls._fetch_file(url)


class GhostbinDownloader(BasicDownloader):
    """"
    Support for ghostbin.com service

    Rewrites links to ghostbin to use raw files.
    """

    @classmethod
    def match(cls, url: str):
        return url.startswith('https://ghostbin.com/paste/')

    @classmethod
    def download(cls, url: str) -> str:
        # Use native raw support fo ghostbin
        if not url.endswith('/raw'):
            url += '/raw'

        return super().download(url)
